@@ -72,6 +72,9 @@ unsigned long int strtoul ();
 		     (c) >= 'A' && (c) <= 'F' ? (c) - 'A' + 10 : (c) - '0')
 #define octtobin(c) ((c) - '0')
 
+/* A value for field_width or precision that indicates it was not specified.  */
+#define UNSPECIFIED INT_MIN
+
 /* The value to return to the calling program.  */
 static int exit_status;
 
@@ -240,7 +243,7 @@ print_esc (const char *escstart)
   int esc_value = 0;		/* Value of \nnn escape. */
   int esc_length;		/* Length of \nnn escape. */
 
-  if (*p == 'x')
+  if (!posixly_correct && *p == 'x')
     {
       /* A hexadecimal \xhh escape sequence must have 1 or 2 hex. digits.  */
       for (esc_length = 0, ++p;
@@ -261,7 +264,7 @@ print_esc (const char *escstart)
 	esc_value = esc_value * 8 + octtobin (*p);
       putchar (esc_value);
     }
-  else if (strchr ("\"\\abcfnrtv", *p))
+  else if (*p && strchr ("\"\\abcfnrtv", *p))
     print_esc_char (*p++);
   else if (!posixly_correct && (*p == 'u' || *p == 'U'))
     {
@@ -292,7 +295,14 @@ print_esc (const char *escstart)
       print_unicode_char (stdout, uni_value, 0);
     }
   else
-    error (EXIT_FAILURE, 0, _("\\%c: invalid escape"), *p);
+    {
+      putchar ('\\');
+      if (*p)
+	{
+	  putchar (*p);
+	  p++;
+	}
+    }
   return p - escstart - 1;
 }
 
@@ -310,7 +320,7 @@ print_esc_string (const char *str)
 
 /* Output a % directive.  START is the start of the directive,
    LENGTH is its length, and ARGUMENT is its argument.
-   If FIELD_WIDTH or PRECISION is non-negative, they are args for
+   If FIELD_WIDTH or PRECISION is UNSPECIFIED, they are args for
    '*' values in those fields. */
 
 static void
@@ -327,16 +337,16 @@ print_direc (const char *start, size_t length, int field_width,
     {
     case 'd':
     case 'i':
-      if (field_width < 0)
+      if (field_width == UNSPECIFIED)
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, xstrtol (argument));
 	  else
 	    printf (p, precision, xstrtol (argument));
 	}
       else
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, field_width, xstrtol (argument));
 	  else
 	    printf (p, field_width, precision, xstrtol (argument));
@@ -347,16 +357,16 @@ print_direc (const char *start, size_t length, int field_width,
     case 'u':
     case 'x':
     case 'X':
-      if (field_width < 0)
+      if (field_width == UNSPECIFIED)
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, xstrtoul (argument));
 	  else
 	    printf (p, precision, xstrtoul (argument));
 	}
       else
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, field_width, xstrtoul (argument));
 	  else
 	    printf (p, field_width, precision, xstrtoul (argument));
@@ -368,16 +378,16 @@ print_direc (const char *start, size_t length, int field_width,
     case 'E':
     case 'g':
     case 'G':
-      if (field_width < 0)
+      if (field_width == UNSPECIFIED)
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, xstrtod (argument));
 	  else
 	    printf (p, precision, xstrtod (argument));
 	}
       else
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, field_width, xstrtod (argument));
 	  else
 	    printf (p, field_width, precision, xstrtod (argument));
@@ -389,16 +399,16 @@ print_direc (const char *start, size_t length, int field_width,
       break;
 
     case 's':
-      if (field_width < 0)
+      if (field_width == UNSPECIFIED)
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, argument);
 	  else
 	    printf (p, precision, argument);
 	}
       else
 	{
-	  if (precision < 0)
+	  if (precision == UNSPECIFIED)
 	    printf (p, field_width, argument);
 	  else
 	    printf (p, field_width, precision, argument);
@@ -420,8 +430,8 @@ print_formatted (const char *format, int argc, char **argv)
   const char *f;		/* Pointer into `format'.  */
   const char *direc_start;	/* Start of % directive.  */
   size_t direc_length;		/* Length of % directive.  */
-  int field_width;		/* Arg to first '*', or -1 if none.  */
-  int precision;		/* Arg to second '*', or -1 if none.  */
+  int field_width;		/* Arg to first '*', or UNSPECIFIED if none. */
+  int precision;		/* Arg to second '*', or UNSPECIFIED if none. */
 
   for (f = format; *f; ++f)
     {
@@ -430,7 +440,7 @@ print_formatted (const char *format, int argc, char **argv)
 	case '%':
 	  direc_start = f++;
 	  direc_length = 1;
-	  field_width = precision = -1;
+	  field_width = precision = UNSPECIFIED;
 	  if (*f == '%')
 	    {
 	      putchar ('%');
@@ -446,7 +456,7 @@ print_formatted (const char *format, int argc, char **argv)
 		}
 	      break;
 	    }
-	  if (strchr ("-+ #", *f))
+	  while (*f == ' ' || *f == '#' || *f == '+' || *f == '-')
 	    {
 	      ++f;
 	      ++direc_length;
@@ -458,6 +468,9 @@ print_formatted (const char *format, int argc, char **argv)
 	      if (argc > 0)
 		{
 		  field_width = xstrtoul (*argv);
+		  if (field_width == UNSPECIFIED)
+		    error (EXIT_FAILURE, 0, _("invalid field width: %s"),
+			   *argv);
 		  ++argv;
 		  --argc;
 		}
@@ -481,6 +494,9 @@ print_formatted (const char *format, int argc, char **argv)
 		  if (argc > 0)
 		    {
 		      precision = xstrtoul (*argv);
+		      if (precision == UNSPECIFIED)
+			error (EXIT_FAILURE, 0, _("invalid precision: %s"),
+			       *argv);
 		      ++argv;
 		      --argc;
 		    }
@@ -499,7 +515,7 @@ print_formatted (const char *format, int argc, char **argv)
 	      ++f;
 	      ++direc_length;
 	    }
-	  if (!strchr ("diouxXfeEgGcs", *f))
+	  if (! (*f && strchr ("diouxXfeEgGcs", *f)))
 	    error (EXIT_FAILURE, 0, _("%%%c: invalid directive"), *f);
 	  ++direc_length;
 	  if (argc > 0)
